@@ -1,12 +1,26 @@
+/*
+    Cparse.java -- Racc Runtime Core for JRuby
+    
+    Copyright (c) 2016 Charles Oliver Nutter <headius@headius.com>
+    
+    Ported from and distributed under the same licence as cparse.c
+    
+    cparse.c -- Racc Runtime Core
+  
+    Copyright (c) 1999-2006 Minero Aoki
+  
+    This library is free software.
+    You can distribute/modify this program under the same terms of ruby.
+*/
 package com.headius.racc;
 
 import org.jruby.Ruby;
 import org.jruby.RubyArray;
 import org.jruby.RubyBasicObject;
 import org.jruby.RubyClass;
-import org.jruby.RubyContinuation;
 import org.jruby.RubyFixnum;
 import org.jruby.RubyHash;
+import org.jruby.RubyKernel;
 import org.jruby.RubyModule;
 import org.jruby.RubyNumeric;
 import org.jruby.RubyObject;
@@ -24,13 +38,14 @@ import org.jruby.runtime.CallSite;
 import org.jruby.runtime.Helpers;
 import org.jruby.runtime.MethodIndex;
 import org.jruby.runtime.ObjectAllocator;
+import org.jruby.runtime.Signature;
 import org.jruby.runtime.ThreadContext;
 import org.jruby.runtime.Visibility;
 import org.jruby.runtime.builtin.IRubyObject;
 import org.jruby.runtime.load.Library;
 
 public class Cparse implements Library {
-    public static final String RACC_VERSION = "1.4.14"; // TODO: parse from Cparse.c
+    public static final String RACC_VERSION = "1.4.15"; // TODO: parse from Cparse.c
 
     public enum TokenType {
         DEFAULT(-1),
@@ -233,10 +248,6 @@ public class Cparse implements Library {
             shift(context, act, tok, val);
         }
 
-        private int REDUCE(ThreadContext context, int act) {
-            return reduce(context, act);
-        }
-
         public void parse_main(ThreadContext context, IRubyObject tok, IRubyObject val, boolean resume) {
             Ruby runtime = context.runtime;
 
@@ -267,7 +278,7 @@ public class Cparse implements Library {
                         i = assert_integer(tmp);
 
                         D_printf("read_next=%d\n", read_next);
-                        if (read_next && (this.t != vFINAL_TOKEN)) {
+                        if (read_next && (!this.t.equals(vFINAL_TOKEN))) {
                             if (this.lex_is_iterator) {
                                 D_puts("resuming...");
                                 if (this.fin != 0) throw runtime.newArgumentError("token given after EOF");
@@ -354,7 +365,21 @@ public class Cparse implements Library {
                         }
                         else if (act < 0 && act > -(this.reduce_n)) {
                             D_puts("reduce");
-                            REDUCE(context, act);
+                            { // macro REDUCE
+                                switch (reduce(context, act)) {
+                                    case 0: /* normal */
+                                        break;
+                                    case 1: /* yyerror */
+                                        branch = USER_YYERROR;
+                                        continue BRANCH;
+                                    case 2: /* yyaccept */
+                                        D_puts("u accept");
+                                        branch = ACCEPT;
+                                        continue BRANCH;
+                                    default:
+                                        break;
+                                }
+                            }
                         }
                         else if (act == -(this.reduce_n)) {
                             branch = ERROR; continue BRANCH;
@@ -367,6 +392,7 @@ public class Cparse implements Library {
                             throw runtime.newRaiseException(RaccBug, "[Cparse Bug] unknown act value " + act);
                         }
 
+                        // fall through
                     case ERROR_RECOVERED:
 
                         if (this.debug) {
@@ -389,9 +415,10 @@ public class Cparse implements Library {
                             call_onerror.call(context, this.parser, this.parser, this.t, val, this.vstack);
                         }
 
+                        // fall through
                     case USER_YYERROR:
                         if (this.errstatus == 3) {
-                            if (this.t == vFINAL_TOKEN) {
+                            if (this.t.equals(vFINAL_TOKEN)) {
                                 this.retval = runtime.getFalse();
                                 this.fin = CP_FIN_EOT;
                                 return;
@@ -464,7 +491,21 @@ public class Cparse implements Library {
                         }
                         else if (act < 0 && act > -(this.reduce_n)) {
                             D_puts("e reduce");
-                            REDUCE(context, act);
+                            { // macro REDUCE
+                                switch (reduce(context, act)) {
+                                    case 0: /* normal */
+                                        break;
+                                    case 1: /* yyerror */
+                                        branch = USER_YYERROR;
+                                        continue BRANCH;
+                                    case 2: /* yyaccept */
+                                        D_puts("u accept");
+                                        branch = ACCEPT;
+                                        continue BRANCH;
+                                    default:
+                                        break;
+                                }
+                            }
                         }
                         else if (act == this.shift_n) {
                             D_puts("e accept");
@@ -489,17 +530,18 @@ public class Cparse implements Library {
         }
 
         private int reduce(ThreadContext context, int act) {
-            IRubyObject code;
             ruleno = -act * 3;
             IRubyObject tag = context.runtime.newSymbol("racc_jump");
-            RubyContinuation rbContinuation = new RubyContinuation(context.runtime, context.runtime.newSymbol("racc_jump"));
-            try {
-                context.pushCatch(rbContinuation.getContinuation());
-                code = reduce0(context);
-                errstatus = assert_integer(parser.getInstanceVariable(ID_ERRSTATUS));
-            } finally {
-                context.popCatch();
-            }
+            IRubyObject code = RubyKernel.rbCatch(context, this,
+                    tag,
+                    CallBlock19.newCallClosure(this, getMetaClass(), Signature.NO_ARGUMENTS,
+                            new BlockCallback() {
+                                @Override
+                                public IRubyObject call(ThreadContext context, IRubyObject[] args, Block block) {
+                                    return reduce0(context);
+                                }
+                            }, context));
+            errstatus = assert_integer(parser.getInstanceVariable(ID_ERRSTATUS));
             return assert_integer(code);
         }
 
@@ -527,13 +569,13 @@ public class Cparse implements Library {
                         /* call action */
                         if (len == 0) {
                             tmp = context.nil;
-                            if (mid != sym_noreduce)
+                            if (!mid.equals(sym_noreduce))
                                 tmp_v = runtime.newArray();
                             if (this.debug)
                                 tmp_t = runtime.newArray();
                         }
                         else {
-                            if (mid != sym_noreduce) {
+                            if (!mid.equals(sym_noreduce)) {
                                 tmp_v = GET_TAIL(context, this.vstack, len);
                                 tmp = ((RubyArray)tmp_v).entry(0);
                             }
@@ -547,7 +589,7 @@ public class Cparse implements Library {
                             }
                             CUT_TAIL(context, this.state, len);
                         }
-                        if (mid != sym_noreduce) {
+                        if (!mid.equals(sym_noreduce)) {
                             if (this.use_result_var) {
                                 tmp = Helpers.invoke(context, this.parser, mid.toString(), tmp_v, this.vstack, tmp);
                             }
@@ -590,7 +632,7 @@ public class Cparse implements Library {
                             D_puts("(goto) check[i] == nil");
                             branch = NOTFOUND; continue BRANCH;
                         }
-                        if (tmp != runtime.newFixnum(k1)) {
+                        if (!tmp.equals(runtime.newFixnum(k1))) {
                             D_puts("(goto) check[i] != table[i]");
                             branch = NOTFOUND; continue BRANCH;
                         }
@@ -741,7 +783,7 @@ public class Cparse implements Library {
         private void call_lexer(ThreadContext context, final CparseParams v) {
             final LexerUnroll lexerUnroll = new LexerUnroll();
             try {
-                v.call_lexmid.call(context, v.lexer, v.lexer, CallBlock19.newCallClosure(v, v.getMetaClass(), Arity.ONE_ARGUMENT, new BlockCallback() {
+                v.call_lexmid.call(context, v.lexer, v.lexer, CallBlock19.newCallClosure(v, v.getMetaClass(), Signature.ONE_ARGUMENT, new BlockCallback() {
                     @Override
                     public IRubyObject call(ThreadContext context, IRubyObject[] args, Block block) {
                         Ruby runtime = context.getRuntime();
@@ -759,7 +801,7 @@ public class Cparse implements Library {
                     }
                 }, context));
             } catch (LexerUnroll maybeOurs) {
-                if (maybeOurs == lexerUnroll) {
+                if (maybeOurs.equals(lexerUnroll)) {
                     return;
                 }
             }
